import csv
import json
from datetime import datetime
from typing import Dict, List

import xmltodict
from dateutil import parser
from rb.cna.cna_graph import CnaGraph
from rb.core.cscl.community import Community
from rb.core.cscl.contribution import Contribution
from rb.core.cscl.conversation import Conversation
from rb.core.cscl.cna_indices_enum import CNAIndices
from rb.core.cscl.participant import Participant
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.core.text_element_type import TextElementType
from rb.processings.cscl.community_processing import (
    compute_sna_metrics, determine_participant_contributions,
    determine_participation)
from rb.processings.cscl.participant_evaluation import (evaluate_interaction,
                                                        evaluate_involvement,
                                                        evaluate_textual_complexity,
                                                        perform_sna)
from rb.similarity.vector_model import VectorModelType
from rb.similarity.vector_model_factory import create_vector_model
from rb.utils.rblogger import Logger

logger = Logger.get_logger()

CONTRIBUTIONS_KEY = 'contributions'
ID_KEY = 'id'
PARENT_ID_KEY = 'parent_id'
TIMESTAMP_KEY = 'timestamp'
TEXT_KEY = 'text'
USER_KEY = 'user'

JSONS_PATH = './jsons/'

def get_json_from_json_file(filename: str) -> Dict:
    conversation_thread = dict()
    contribution_list = []

    with open(filename, "rt", encoding='utf-8') as json_file:
        contribution_list_json = json.load(json_file)
        for cnt in contribution_list_json:
            contribution = {
                ID_KEY: cnt['genid'],
                PARENT_ID_KEY: cnt["ref"],
                TIMESTAMP_KEY: int(float(cnt['time'])),
                USER_KEY: cnt["nickname"],
                TEXT_KEY: cnt["text"],
            }
            contribution_list.append(contribution)

    conversation_thread[CONTRIBUTIONS_KEY] = contribution_list

    return conversation_thread

def get_json_from_csv(filename: str) -> Dict:
    conversation_thread = dict()
    contribution_list = []

    with open(filename) as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=',')
        first = True			

        for row in csv_reader:
            if first:
                first = False
            else:
                contribution = dict()

                contribution[ID_KEY] = row[0]
                contribution[PARENT_ID_KEY] = row[1]
                contribution[USER_KEY] = row[2]
                contribution[TEXT_KEY] = row[3]
                contribution[TIMESTAMP_KEY] = row[4]

                contribution_list.append(contribution)

    conversation_thread[CONTRIBUTIONS_KEY] = contribution_list

    return conversation_thread

def parse_large_csv(filename: str) -> Dict:
    conversation_thread = dict()
    contribution_list = []

    with open(filename) as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=',')
        first = True			

        for row in csv_reader:
            if first:
                first = False
            else:
                contribution = dict()

                contribution[ID_KEY] = row[0]
                contribution[PARENT_ID_KEY] = row[1]
                if row[1] == '':
                    contribution[PARENT_ID_KEY] = '-1'

                contribution[USER_KEY] = row[2]
                contribution[TEXT_KEY] = row[6]

                contribution[TIMESTAMP_KEY] = datetime.timestamp(parser.parse(row[3], ignoretz=True, fuzzy=True))

                contribution_list.append(contribution)

    conversation_thread[CONTRIBUTIONS_KEY] = contribution_list

    return conversation_thread

def load_from_xml(lang: Lang, filename: str) -> Dict:
	with open(filename, "rt") as f:
		try:
			my_dict=xmltodict.parse(f.read())
		except:
			print(filename)
			raise
		turns = my_dict["Dialog"]["Body"]["Turn"]
		if not isinstance(turns, List):
			turns = [turns]
		contributions = [
			{
				ID_KEY: int(utterance["@genid"]) - 1,
				PARENT_ID_KEY: int(utterance["@ref"]) - 1,
				TIMESTAMP_KEY: datetime.timestamp(datetime.strptime(utterance["@time"],'%Y-%m-%d %H:%M:%S.%f')),
				USER_KEY: turn["@nickname"],
				TEXT_KEY: utterance["#text"],
			}
			for turn in turns
			for utterance in (turn["Utterance"] if isinstance(turn["Utterance"], List) else [turn["Utterance"]])
		]

		return {CONTRIBUTIONS_KEY: contributions}


def export_individual_statistics(participants: List[Participant], filename: str):
	first = True

	cscl_keys = [CNAIndices.CONTRIBUTIONS_SCORE, CNAIndices.SOCIAL_KB, CNAIndices.NO_CONTRIBUTION, CNAIndices.OUTDEGREE, CNAIndices.INDEGREE,
				CNAIndices.NO_NEW_THREADS, CNAIndices.NEW_THREADS_OVERALL_SCORE, CNAIndices.NEW_THREADS_CUMULATIVE_SOCIAL_KB,
				CNAIndices.AVERAGE_LENGTH_NEW_THREADS]

	evaluate_interaction(conv)
	evaluate_involvement(conv)
	evaluate_textual_complexity(conv)
	perform_sna(conv, False)
	with open(filename, 'w') as f:
		for p in participants:
			indices = p.indices
			keys = ['participant'] + [str(k.value) for k in CNAIndices]
			values = [p.get_id()] + [str(p.get_index(k)) for k in cscl_keys]

			if first:
				f.write(','.join(keys) + '\n')
				first = False

			f.write(','.join(values) + '\n')
		f.flush()
		f.close()

def export_textual_complexity(participants: List[Participant], filename: str):
	first = True

	with open(filename, 'w') as f:
		for p in participants:
			indices = p.textual_complexity_indices
			keys = ['participant'] + [str(k) for k in indices.keys()]
			values = [p.get_id()] + [str(indices[k]) for k in indices.keys()]

			if first:
				f.write(','.join(keys) + '\n')
				first = False

			f.write(','.join(values) + '\n')
		f.flush()
		f.close()

def test_community_processing():
	print('Testing Community Processing')

	conv_thread = []
	for i in range(1, 20):
		conversation = get_json_from_json_file("./jsons/conversation_" + str(i) + ".json")
		conv_thread.append(conversation)

	community = Community(lang=Lang.EN, container=None, community=conv_thread)
	en_coca_word2vec = create_vector_model(Lang.EN, VectorModelType.from_str("word2vec"), "coca")
	community.graph = CnaGraph(docs=[community], models=[en_coca_word2vec])

	participant_list = community.get_participants()
	names = list(map(lambda p: p.get_id(), participant_list))

	print('Participants are:')
	print(names)

	print('Begin computing indices')

	determine_participant_contributions(community)
	determine_participation(community)
	compute_sna_metrics(community)
	# determine_textual_complexity(community)

	print('Finished computing indices')

	for p in participant_list:
		print('Printing for participant ' + p.get_id())

		print(p.get_index(CNAIndices.CONTRIBUTIONS_SCORE))
		print(p.get_index(CNAIndices.NO_CONTRIBUTION))
		print(p.get_index(CNAIndices.SOCIAL_KB))
		print(p.get_index(CNAIndices.INDEGREE))
		print(p.get_index(CNAIndices.OUTDEGREE))
		print(p.get_index(CNAIndices.NO_NEW_THREADS))
		print(p.get_index(CNAIndices.NEW_THREADS_OVERALL_SCORE))
		print(p.get_index(CNAIndices.NEW_THREADS_CUMULATIVE_SOCIAL_KB))
		print(p.get_index(CNAIndices.AVERAGE_LENGTH_NEW_THREADS))

		# print('Printing textual complexity for participant ' + p.get_id())
		# indices = p.textual_complexity_indices
		#
		# for key, value in indices.items():
		# 	print('Index ' + key + ' is ' + str(value))

		print('---------------------')

	for n1 in names:
		for n2 in names:
			print('Score for ' + n1 + ' ' + n2 + ' is:')
			print(community.get_score(n1, n2))

	export_individual_statistics(community.get_participants(), './individualStats.csv')
	# export_textual_complexity(community.get_participants(), './textualComplexity.csv')

def test_participant_evaluation():
	print('Testing English CSV')

	conv_thread = []
	for i in range(1, count_jsons + 1):
		conversation = get_json_from_json_file(JSONS_PATH + "conversation_" + str(i) + ".json")
		conv_thread.append(conversation)

	community = Community(lang=Lang.EN, container=None, community=conv_thread)
	en_coca_word2vec = create_vector_model(Lang.EN, VectorModelType.from_str("word2vec"), "coca")
	community.graph = CnaGraph(docs=[community], models=[en_coca_word2vec])

	conv = community.get_conversations()[0]

	# participant_list = conv.get_participants()
	participant_list = community.get_participants()
	names = list(map(lambda p: p.get_id(), participant_list))

	print('Participants are:')
	print(names)

	conv_thread = parse_large_csv('./thread.csv')
	print('Begin computing indices')

	evaluate_interaction(conv)
	evaluate_involvement(conv)
	# evaluate_textual_complexity(conv)
	perform_sna(conv, False)

	print('Finished computing indices')

	for p in participant_list:
		print('Printing for participant ' + p.get_id())

		print(p.get_index(CNAIndices.CONTRIBUTIONS_SCORE))
		print(p.get_index(CNAIndices.NO_CONTRIBUTION))
		print(p.get_index(CNAIndices.SOCIAL_KB))
		print(p.get_index(CNAIndices.INDEGREE))
		print(p.get_index(CNAIndices.OUTDEGREE))

		# print('Printing textual complexity for participant ' + p.get_id())
		# indices = p.textual_complexity_indices
		#
		# for key, value in indices.items():
		# 	print('Index ' + key + ' is ' + str(value))

		print('---------------------')

	for n1 in names:
		for n2 in names:
			print('Score for ' + n1 + ' ' + n2 + ' is:')
			print(conv.get_score(n1, n2))

	export_individual_statistics(conv.get_participants(), './individualStats.csv')

def main():
	# Test community processing - English discussion csv
	test_community_processing()

	# Test participant evaluation - English discussion csv
	#test_participant_evaluation()

	# Test for French discussion XML

	'''

	print('Testing French XML')

	conv_thread = CsvParser.load_from_xml(Lang.FR, './conpa-MEEF-anonyme.xml')
	
	community = Community(lang=Lang.FR, container=None, community=[conv_thread])
	fr_coca_word2vec = create_vector_model(Lang.FR, VectorModelType.from_str("word2vec"), "coca")
	community.graph = CnaGraph(docs=[community], models=[fr_coca_word2vec])

	conv = community.get_conversations()[0]

	compute_indices(conv)

	'''

if __name__ == '__main__':
	main()



from typing import Dict, List, Union

import networkx as nx
from networkx.algorithms.centrality import betweenness_centrality, closeness_centrality, eigenvector_centrality
from rb.cna.cna_graph import CnaGraph
from rb.complexity.complexity_index import compute_indices
from rb.core.block import Block
from rb.core.cscl.cna_indices_enum import CNAIndices
from rb.core.cscl.contribution import Contribution
from rb.core.cscl.conversation import Conversation
from rb.core.cscl.community import Community
from rb.core.lang import Lang
from rb.core.pos import POS
from rb.core.text_element import TextElement
from rb.utils.rblogger import Logger


def get_block_importance(graph: nx.DiGraph, a: Block, b: Block) -> float:
	return graph.edges[a, b]["weight"] if (a, b) in graph.edges else 0.


def evaluate_interaction(conversation: Union[Conversation, Community]):
	cna_graph = conversation.container.graph if conversation.container is not None else conversation.graph
	
	participants = conversation.get_participants()
	if isinstance(conversation, Community):
		contributions = [contribution for conv in conversation.components for contribution in conv.get_contributions()]
	else:
		contributions = conversation.get_contributions()
	conversation.init_scores()

	if (len(participants) == 0):
		return

	for i, contribution1 in enumerate(contributions):
		p1 = contribution1.get_participant().get_id()

		for contribution2 in contributions[:i]:
			weight = get_block_importance(cna_graph.filtered_graph, contribution1, contribution2)
			if weight > 0:
				p2 = contribution2.get_participant().get_id()
				conversation.update_score(p1, p2, weight)


def evaluate_involvement(conversation: Union[Conversation, Community]):
	cna_graph = conversation.container.graph if conversation.container is not None else conversation.graph
	participants = conversation.get_participants()

	if (len(participants) == 0):
		return
	if isinstance(conversation, Community):
		contributions = [contribution for conv in conversation.components for contribution in conv.get_contributions()]
	else:
		contributions = conversation.get_contributions()
	
	for i, contribution in enumerate(contributions):
		p = contribution.get_participant()

		current_value = p.get_index(CNAIndices.CONTRIBUTIONS_SCORE)
		p.set_index(CNAIndices.CONTRIBUTIONS_SCORE, current_value + cna_graph.importance[contribution])

		added_value = sum([
            get_block_importance(cna_graph.filtered_graph, contribution, prev)
            for j, prev in enumerate(contributions[:i])
            if p != prev.get_participant()
        ])
		p.set_index(CNAIndices.SOCIAL_KB, p.get_index(CNAIndices.SOCIAL_KB) + added_value)

def evaluate_textual_complexity(conversation: Conversation):
	participants = conversation.get_participants()
	
	for p in participants:
		contributions = conversation.get_participant_contributions(p.get_id())
		p.set_eligible_contributions(contributions)

		Conversation.create_participant_conversation(conversation.lang, p)

		own_conversation = p.own_conversation

		print('Begin computing textual complexity')
		compute_indices(doc=own_conversation)
		print('Finished computing textual complexity')

		for key, value in own_conversation.indices.items():
			p.set_textual_index(repr(key), value)


def perform_sna(conversation: Conversation, needs_anonymization: bool) -> nx.DiGraph:
    participants = conversation.get_participants()
    cna_graph = conversation.container.graph if conversation.container is not None else conversation.graph
    participants_graph = nx.DiGraph()
    participants_graph.add_nodes_from(participants)
    for u, v, weight in cna_graph.filtered_graph.edges(data="weight"):
        if isinstance(u, Contribution) and isinstance(v, Contribution) and u.get_participant() != v.get_participant() and u.timestamp < v.timestamp:
            if participants_graph.has_edge(u.get_participant(), v.get_participant()):
                participants_graph[u.get_participant()][v.get_participant()]["weight"] += weight
            else:
                participants_graph.add_edge(u.get_participant(), v.get_participant(), weight=weight)
    betweeness = betweenness_centrality(participants_graph)
    closeness = closeness_centrality(participants_graph)
    eigen = eigenvector_centrality(participants_graph)
    for p in participants:
        p.set_index(CNAIndices.OUTDEGREE, participants_graph.out_degree(p, weight="weight"))
        p.set_index(CNAIndices.INDEGREE, participants_graph.in_degree(p, weight="weight"))
        p.set_index(CNAIndices.BETWEENNESS, betweeness[p])
        p.set_index(CNAIndices.CLOSENESS, closeness[p])
        p.set_index(CNAIndices.EIGENVECTOR, eigen[p])
        p.set_index(CNAIndices.INTERACTION_SCORE, sum([participants_graph[p][n]["weight"] for n in participants_graph.neighbors(p)]))
    return participants_graph



from typing import Dict, List, Tuple

from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.core.text_element_type import TextElementType
from rb.core.cscl.contribution import Contribution
from rb.core.cscl.conversation import Conversation
from rb.core.cscl.participant import Participant
from rb.similarity.vector_model_factory import create_vector_model
from rb.similarity.vector_model import VectorModelType

from rb.utils.rblogger import Logger

logger = Logger.get_logger()

# muchii intre replici parinte-copil?
# muchii intre replici ale aceluiasi user ?

class Community(TextElement):

	'''
		community -> list of JSON objects (crawled from Reddit), with
					 the same structure as described in Conversation class
	'''
	def __init__(self, lang: Lang, 
				 community: List[Dict],
				 container: TextElement = None,
				 depth: int = TextElementType.COMM.value,
				 start_date: int = None,
				 end_date: int = None):

		TextElement.__init__(self, lang=lang, text=None,
							 depth=depth, container=container)

		self.components = []
		self.start_date = start_date
		self.end_date = end_date

		self.participant_map = dict()

		for conversation_thread in community:
			current_conversation = Conversation(lang, container=self,
						 						conversation_thread=conversation_thread)

			self.components.append(current_conversation)

		self.participants = self.union_participants()
		self.participant_contributions = self.union_contributions()
		self.first_contribution_date, self.last_contribution_date = self.find_contribution_range()

		self.graph = None

		self.eligible_contributions = []
		self.timeframe_subcommunities = []
		self.scores = dict()
		self.init_scores()

		self.community_raw = community

	def union_participants(self) -> List[Participant]:
		return [self.participant_map[participant_id] for participant_id in self.participant_map]

	def union_contributions(self) -> Dict:
		contributions = dict()

		for conversation in self.components:
			for participant in conversation.get_participants():
				if not (participant in contributions):
					contributions[participant.get_id()] = []

				contributions[participant.get_id()] += conversation.get_participant_contributions(participant.get_id())

		return contributions

	def get_community_raw(self) -> List[Dict]:
		return self.community_raw

	def find_contribution_range(self) -> Tuple[int, int]:
		first_contribution = None
		last_contribution = None

		for conversation in self.components:
			for contribution in conversation.get_contributions():
				timestamp = contribution.get_timestamp()

				if self.is_eligible(timestamp):
					if first_contribution == None or timestamp < first_contribution:
						first_contribution = timestamp

					if last_contribution == None or timestamp > last_contribution:
						last_contribution = timestamp

		if self.start_date is None:
			self.start_date = first_contribution
		if self.end_date is None:
			self.end_date = last_contribution

		return first_contribution, last_contribution

	def is_eligible(self, timestamp: int) -> bool:
		if self.start_date != None and self.end_date != None:
			return (timestamp >= self.start_date and timestamp <= self.end_date)
		
		return True

	def add_eligible_contribution(self, contribution: Contribution):
		self.eligible_contributions.append(contribution)

	def add_subcommunity(self, community):
		self.timeframe_subcommunities.append(community)

	def get_conversations(self) -> List[Conversation]:
		return self.components

	def get_participant(self, participant_id: str) -> Participant:
		return self.participant_map[participant_id]

	def get_participants(self) -> List[Participant]:
		return self.participants

	def get_participant_contributions(self, participant_id: str) -> List[Contribution]:
		return self.participant_contributions[participant_id]

	def get_first_contribution_date(self) -> int:
		return self.first_contribution_date

	def get_last_contribution_date(self) -> int:
		return self.last_contribution_date

	def init_scores(self):
		for a in self.participants:
			self.scores[a.get_id()] = dict()

			for b in self.participants:
				self.scores[a.get_id()][b.get_id()] = 0

	def get_score(self, a: str, b: str) -> float:
		return self.scores[a][b]

	def update_score(self, a: str, b: str, value: float):
		self.scores[a][b] += value
